/*
 * Copyright (C) 2009-2010 SENJU Jiro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <sys/epoll.h>
#include <sys/wait.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "lib.h"
#include "sample_sub.h"

static int *sock_child, dummy_stdout;

static pid_t do_dummy_wait(pid_t pid, int opt)
{
	int status;

	pid = waitpid(pid, &status, opt);
	assert((opt & WNOHANG) || pid > 0);
	if (pid > 0) {
		assert(WIFEXITED(status));
		//assert(!WEXITSTATUS(status));
	}

	return pid;
}

int dummy_wait(int sigfd, int epfd)
{
	int err;
	pid_t pid;
	struct epoll_event ev;
	struct signalfd_siginfo ssi;
	ssize_t ssz;

	err = 0;
	ssz = read(sigfd, &ssi, sizeof(ssi));
	assert(ssz == sizeof(ssi));
	assert(ssi.ssi_signo == SIGCHLD);

	ev.events = EPOLLOUT /* | EPOLLET */;
	while (1) {
		pid = do_dummy_wait(-1, WNOHANG);
		if (pid <= 0)
			break;

		ev.data.fd = sock_child[pid];
		err = epoll_ctl(epfd, EPOLL_CTL_ADD, ev.data.fd, &ev);
		assert(!err);
	}

	return err;
}

int dummy_busy(int sock, int do_wait)
{
	int err;
	pid_t pid;

	err = 0;
	pid = fork();
	if (pid > 0) {
		/* parent */
		if (!do_wait)
			sock_child[pid] = sock;
		else
			do_dummy_wait(pid, !WNOHANG);
	} else if (!pid) {
		/* child */
		char *argv[] = {
			"find", "/tmp",
			NULL
		};

		err = dup2(dummy_stdout, STDOUT_FILENO);
		assert(err == STDOUT_FILENO);
		err = dup2(dummy_stdout, STDERR_FILENO);
		assert(err == STDERR_FILENO);
		execv("/usr/bin/find", argv);
	} else {
		assert(0);
		err = -1;
	}

	return err;
}

static int dummy_sigfd(int epfd)
{
	int err;
	sigset_t mask;
	struct epoll_event ev;

	sigemptyset(&mask);
	sigaddset(&mask, SIGCHLD);
	err = sigprocmask(SIG_BLOCK, &mask, NULL);
	assert(!err);
	ev.data.fd = signalfd(-1, &mask, /*flags unused*/0);
	assert(ev.data.fd >= 0);
	err = set_nonblock(ev.data.fd, 1);
	assert(!err);

	ev.events = EPOLLIN /* | EPOLLET */;
	err = epoll_ctl(epfd, EPOLL_CTL_ADD, ev.data.fd, &ev);
	assert(!err);

	return ev.data.fd;
}

int dummy_init(int epfd)
{
	int fd, pid_max;
	char a[16];
	ssize_t ssz;

	assert(WNOHANG);
	dummy_stdout = open("/dev/null", O_RDWR);
	assert(dummy_stdout >= 0);

	if (epfd >= 0) {
		fd = open("/proc/sys/kernel/pid_max", O_RDONLY);
		assert(fd >= 0);
		ssz = read(fd, a, sizeof(a));
		close(fd);
		assert(ssz > 0);
		a[ssz - 1] = 0;
		pid_max = atoi(a);
		assert(pid_max > 0);
		//assert(pid_max <= 0x8000); /* default */

		sock_child = calloc(pid_max, sizeof(*sock_child));
		assert(sock_child);
		/*
		 * fake index origin. it must be safe since its lower range is
		 * reserved.
		 */
		sock_child--;

		fd = dummy_sigfd(epfd);
	}

	return fd;
}

/* ---------------------------------------------------------------------- */

static char **path;

int do_read(int sock, int epfd)
{
	int err;
	ssize_t ssz;
	char a[BUFSIZ];

	ssz = read(sock, a, sizeof(a));
	assert(ssz > 0);
	assert(ssz != sizeof(a));

	path[sock] = get_path(a);
	assert(path[sock]);
	path[sock] = strdup(path[sock]);
	assert(path[sock]);

	err = epoll_ctl(epfd, EPOLL_CTL_DEL, sock, NULL);
	assert(!err);

	dummy_busy(sock, /*do_wait*/0);

	return err;
}

int do_close(int sock, int epfd)
{
	int err;

	err = epoll_ctl(epfd, EPOLL_CTL_DEL, sock, NULL);
	assert(!err);
	err = close(sock);
	assert(!err);
	free(path[sock]);
	path[sock] = NULL;

	return err;
}

/* sendfile and close */
int do_sfc(int sock, int epfd)
{
	assert(path[sock]);
	do_sendfile(sock, path[sock]);
	return do_close(sock, epfd);
}

int sample_init(int maxopen)
{
	path = calloc(sizeof(*path), maxopen);
	assert(path);
	return 0;
}
